/*

Copyright (C) 2000 - 2003 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#ifndef __libnd_packet_iterator_h
#define __libnd_packet_iterator_h

#include <libnd_types.h>
#include <libnd_packet.h>
#include <libnd_packet_iterator.h>
#include <libnd_tpm.h>
#include <libnd_tp.h>

typedef enum {

  /* Iterate over selected packets only, in read-only or
   * read-write mode.
   */
  LND_PACKET_IT_SEL_R,
  LND_PACKET_IT_SEL_RW,

  /* Iterate over all packets in currently loaded
   * trace part, read-only or read-write.
   */
  LND_PACKET_IT_PART_R,
  LND_PACKET_IT_PART_RW,
  
  /* Iterate over an area in the trace, starting at
   * currently loaded part and going arbitrarily far.
   * Comes in read-only and read-write flavours.
   */
  LND_PACKET_IT_AREA_R,
  LND_PACKET_IT_AREA_RW,

} LND_PacketIteratorMode;



struct lnd_packet_iterator
{
  /* The trace whose packets are iterated. */
  LND_Trace             *trace;

  LND_PacketIteratorMode mode;
  
  /* Whether to skip filtered packets or not. */
  gboolean               skip_filtered;

  /* Whether to drop the current packet (i.e., not write it out)
   * in LND_PACKET_IT_AREA_RW mode.
   */
  gboolean               drop_current;
  
  /* The packet that's currently iterated. */
  LND_Packet            *current;

  LND_Packet             packet;

  /* End condition for LND_PACKET_IT_AREA_* -- we stop
   * once we reach the given part at the given offset.
   */
  off_t                  offset;

  LND_TraceLoc           stop_loc;
  gboolean               stopped;

  /* A counter for the number of packets we've iterated. */
  guint64                packet_count;

};


/* An iterator interface for iterating over a trace's packets,
   either all of them or only the selected ones. If apply-to-all
   is activated, iterating over all selected packets automatically
   equals iterating over all packets. */

/**
 * libnd_pit_init - packet iterator initializer
 * @pit: iterator to initialize.
 * @trace: trace whose packets are to be iterated.
 *
 * Initializes the packet iterator pointed to by @pit to
 * the iterator mode and possibly the trace area (if the
 * trace's iterator mode is one of the area iterator modes)
 * currently used by @trace.
 *
 * Returns: %TRUE when iterator initialization was
 * successful, %FALSE otherwise.
 */
gboolean            libnd_pit_init(LND_PacketIterator *pit,
				   LND_Trace *trace);

/**
 * libnd_pit_init_mode - initializes iterator using a given mode.
 * @pit: packet iterator to initialize.
 * @trace: the trace that will be iterated over.
 * @mode: the mode to use for iteration.
 *
 * The function prepares @pit to iterate over @trace using @mode.
 * If the mode is a trace area iterator, the trace's current
 * trace area is used.
 *
 * Returns: %TRUE when iterator initialization was
 * successful, %FALSE otherwise.
 */
gboolean            libnd_pit_init_mode(LND_PacketIterator *pit,
					LND_Trace *trace,
					LND_PacketIteratorMode mode);
					
/**
 * libnd_pit_cleanup - releases iterator state.
 * @pit: iterator to clean up.
 *
 * This function releases any state created by @pit. It does not free
 * @pit itself, as packet iterators are conveniently used from the function
 * call stack. You *must* call this function if you abort an iteration
 * prematurely, that is, if you don't just break out of a for-loop when
 * libnd_pit_get() returns %NULL. In this and only in this case is the
 * cleanup performed automatically.
 */
void                libnd_pit_cleanup(LND_PacketIterator *pit);

/**
 * libnd_pit_get - accessor function.
 * @pit: iterator to retrieve current packet from.
 *
 * Accessor function, returns the packet the iterator currently
 * points to.
 *
 * Returns: currently iterated packet.
 */
LND_Packet         *libnd_pit_get(LND_PacketIterator *pit);


/**
 * libnd_pit_next - advances the iterator.
 * @pit: iterator to advance
 *
 * This function advances the iterator, and returns that packet. If there are
 * no more packets, %NULL is returned. Upon advancing the iterator,
 * the progressbar and the packet's tcpdump line automatically
 * get updated.
 *
 * Returns: next packet.
 */
LND_Packet         *libnd_pit_next(LND_PacketIterator *pit);


/**
 * libnd_pit_get_count - get number of iterated packets.
 * @pit: iterator to get current count from.
 *
 * Returns: the number of packets iterated using @pit.
 */
guint64            libnd_pit_get_count(LND_PacketIterator *pit);


/**
 * libnd_pit_skip_filtered - enable/disable skipping of filtered packets.
 * @pit: iterator to adjust.
 * @skip_filtered: whether to skip or not.
 *
 * The function enables or disables skipping of filtered packets for
 * the given iterator. By default, filtered packets are skipped!
 */
void               libnd_pit_skip_filtered(LND_PacketIterator *pit, gboolean skip_filtered);


/**
 * libnd_pit_drop_current - don't write out current packet.
 * @pit: packet iterator.
 *
 * This function only makes sense in iteration mode LND_PACKET_IT_AREA_RW.
 * Calling it during iteration causes the current packet not to be written
 * out. This only applies to the current packet -- the corresponding flag
 * is reset in every iteration.
 */
void               libnd_pit_drop_current(LND_PacketIterator *pit);

#endif
