
/*

Copyright (C) 2000 - 2003 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/param.h>
#include <dirent.h>

#include <libnd_misc.h>
#include <libnd_prefs.h>

struct lnd_plugin
{
  const char *(*name) (void);
  const char *(*author) (void);
  const char *(*version) (void);

  int         (*init) (void);
  int         (*run) (LND_Trace *trace, void *user_data);

  char        *filename;
  lt_dlhandle  lt;
};

static GList *plugins = NULL;  /* list<LND_Plugin*> */

static LND_Plugin  *plugin_new(char *filename);
static void         plugin_scan(const char *dirname);
static void         plugin_hook_up(LND_Plugin *p);
static void         plugin_load(void);
static int          plugin_cmp(LND_Plugin *p1, char *filename);

static const char  *plugin_dummy_author(void) { return "No author(s) specified."; }
static const char  *plugin_dummy_version(void) { return "No version specified."; }
static int          plugin_dummy_init(void) { return 0; }

static int          plugin_dummy_run(LND_Trace *trace, void *user_data)
{
  return 0;
  TOUCH(trace);
  TOUCH(user_data);
}


static LND_Plugin *
plugin_new(char *filename)
{
  LND_Plugin *plugin;

  D_ENTER;

  plugin = (LND_Plugin*) g_new0(LND_Plugin, 1);
  D_ASSERT_PTR(plugin);
  if (!plugin)
    D_RETURN_(NULL);

  if (filename)
    plugin->filename = g_strdup(filename);

  /* We hook in dummy callbacks for segfault protection */
  plugin->author      = plugin_dummy_author;
  plugin->version     = plugin_dummy_version;
  plugin->init        = plugin_dummy_init;
  plugin->run         = plugin_dummy_run;
  
  D_RETURN_(plugin);
}


static void
plugin_free(LND_Plugin *plugin)
{
  D_ENTER;

  if (!plugin)
    D_RETURN;
  
  lt_dlclose(plugin->lt);
  g_free(plugin->filename);
  g_free(plugin);

  D_RETURN;
}


static int    
plugin_cmp(LND_Plugin *p1, char *filename)
{
  if (!p1 || !filename)
    return (-1);

  return (strcmp(p1->filename, filename));
}


static void
plugin_scan(const char *dir)
{
  DIR             *dirp;
  struct dirent   *de;
  char            *dot;

  D(("Scanning '%s' for plugins\n", dir));

  if ((dirp = opendir(dir)) != NULL)
    {
      for (de = readdir(dirp); de; de = readdir(dirp))
	{
	  if (de->d_name[0] != '.')
	    {
	      dot = strrchr(de->d_name, '.');

	      if (dot)
		*dot = '\0';

	      if (!g_list_find_custom(plugins, de->d_name, (GCompareFunc)plugin_cmp))
		plugins = g_list_append(plugins, plugin_new(de->d_name));
	    }
	}      
      closedir(dirp);
    }
}


static void       
plugin_hook_up(LND_Plugin *plugin)
{
  lt_ptr_t     sym;

  if ((sym = lt_dlsym(plugin->lt, "name")) != NULL)
    plugin->name = sym;
  if ((sym = lt_dlsym(plugin->lt, "author")) != NULL)
    plugin->author = sym;
  if ((sym = lt_dlsym(plugin->lt, "version")) != NULL)
    plugin->version = sym;

  if ((sym = lt_dlsym(plugin->lt, "init")) != NULL)
    plugin->init = sym;
  if ((sym = lt_dlsym(plugin->lt, "run")) != NULL)
    plugin->run = sym;

  plugin->init();
}


static gint
plugin_strcmp(gconstpointer a, gconstpointer b)
{
  LND_Plugin *p1 = (LND_Plugin *) a;
  LND_Plugin *p2 = (LND_Plugin *) b;

  return strcmp(p1->name(), p2->name());
}

static void       
plugin_load(void)
{
  GList       *l;
  LND_Plugin   *plugin;
  char         s[MAXPATHLEN];

  for (l = g_list_first(plugins); l; l = g_list_next(l))
    {
      plugin = (LND_Plugin*)l->data;

      g_snprintf(s, MAXPATHLEN, "%s/%s", libnd_prefs_get_plugin_dir_user(), plugin->filename);

      if ((plugin->lt = lt_dlopenext(s)))
	{
	  plugin_hook_up(plugin);
	}
      else
	{
	  g_snprintf(s, MAXPATHLEN, "%s/%s", libnd_prefs_get_plugin_dir_global(), plugin->filename);

	  if ((plugin->lt = lt_dlopenext(s)))
	    {
	      plugin_hook_up(plugin);
	    }
	}
    }

  /* Now sort the list of plugins alphanumerically,
     so that we automatically get sorted menus etc.
  */
  plugins = g_list_sort(plugins, plugin_strcmp);
}


static void
plugin_unload(void)
{
  GList     *l;
  LND_Plugin *pl;

  if (!plugins)
    return;

  for (l = plugins; l; l = g_list_next(l))
    {
      pl = (LND_Plugin *) l->data;

      plugin_free(pl);
      l->data = NULL;
    }

  g_list_free(plugins);
  plugins = NULL;
}


void 
libnd_plugin_init(void)
{
  static gboolean deja_vu = FALSE;

  D_ENTER;

  if (deja_vu)
    plugin_unload();

  if (lt_dlinit() != 0)
    {
      fprintf(stderr, "Could not initialize libltdl -- exiting.\n");
      exit(-1);
    }

  plugin_scan(libnd_prefs_get_plugin_dir_user());
  plugin_scan(libnd_prefs_get_plugin_dir_global());
  plugin_load();

  /* FIXME could add signal here to tell observers that plugins (re)loaded */

  deja_vu = TRUE;
  D_RETURN;
}


LND_Plugin   *
libnd_plugin_find(const char *name)
{
  GList       *l;
  LND_Plugin   *plugin;

  for (l = g_list_first(plugins); l; l = g_list_next(l))
    {
      plugin = (LND_Plugin*)l->data;

      if (!strcmp(plugin->name(), name))
	return plugin;
    }

  D(("No plugin with name '%s' found.\n", name));
  return NULL;
}


void    
libnd_plugin_run(LND_Plugin *plugin,
	      LND_Trace *trace,
	      void *user_data)
{
  plugin->run(trace, user_data);
}



const char *
libnd_plugin_get_name(LND_Plugin *plugin)
{
  if (!plugin)
    return (NULL);
  
  return plugin->name();
}


const char   *
libnd_plugin_get_author(LND_Plugin *plugin)
{
  if (!plugin)
    return (NULL);

  return plugin->author();
}


const char   *
libnd_plugin_get_version(LND_Plugin *plugin)
{
  if (!plugin)
    return (NULL);

  return plugin->version();
}


void           
libnd_plugin_foreach(LND_PluginFunc callback,
		     void *user_data)
{
  GList *l;

  if (!callback)
    return;
  
  for (l = g_list_first(plugins); l; l = g_list_next(l))
    callback((LND_Plugin *)l->data, user_data);
}

